/******************************************************************************
 *
 * Filename: codegen.c
 *
 * Author:   Robert Thornburrow
 * Project:  Cycle - The CYbot Control LanguagE
 * Created:  6th June 2003
 *
 * Purpose:  The Code Generator - the bit that produces the .03p file.
 *
 * (c) Copyright R.H.Thornburrow, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>

#include "cycdefs.h"
#include "errors.h"


/* Numbers of the various blocks */
#define START       0
#define IFTHENELSE  1
#define ARITHMETIC  2
#define DELAY       3
#define SOUND       4
#define MOTORS      5
#define SONAR       6
#define JUMP        7
#define LAND        8
#define ANTENNAE    9
#define LIGHT      10
#define STOP       15

/* Link generation stuff */
#define PUSH_PREV(pid,pport)  { previd[prevc] = pid; prevport[prevc] = pport; prevc++; \
								if( prevc > MAX_DEPTH ) \
								    codegen_error( "out of prev link stack space - increase MAX_DEPTH.", NULL ); }

#define SAVE_PREV(temp)		{ for( ; prevc > 0; prevc-- )	{ \
							  temp##id[temp##c] = previd[prevc-1]; \
							  temp##port[temp##c] = prevport[prevc-1]; temp##c++; } \
							  if( temp##c > MAX_DEPTH ) \
								  codegen_error( "out of " #temp " link stack space - increase MAX_DEPTH.", NULL ); }

#define RESTORE_PREV(temp)  for( ; temp##c > 0; temp##c-- )	{ \
			PUSH_PREV( temp##id[temp##c-1], temp##port[temp##c-1] ); }

#define MAX_DEPTH 100

int id, x, y, landid;
int previd[MAX_DEPTH], prevport[MAX_DEPTH], prevc = 0;
LineList* lines = NULL;


/* Forward reference */
void generateStatementList( FILE* out, StatementList* stmtlist );


const char* getSimpleExpression( char* exprbuf, Expression* expr )
{
	switch( expr->kind )
	{
	case variable_expr:
		return expr->data.variable_part->name;
	case integer_expr:
		sprintf( exprbuf, "%d", expr->data.integer_part );
		return exprbuf;
	case string_expr:
		sprintf( exprbuf, "%s", expr->data.string_part );
		return exprbuf;
	case constant_expr:
		/* Lookup constant value */
		if( expr->data.constant_part.value
			&& (expr->data.constant_part.value->kind == integer_expr) )
			sprintf( exprbuf, "%d", expr->data.constant_part.value->data.integer_part );
		else
			codegen_error( "constant with non-integer value.", expr->lineNo );
		return exprbuf;
	}

	codegen_error( "Illegal expression.", expr->lineNo );
	return NULL;
}


void block0( FILE* out, int kind )
{
	fprintf( out, "==Block\r%d\r%d\r%d\r%d\r0\r", id++, kind, x, y );
	fflush( out );
	x+=30; y+=40;
}


void block1( FILE* out, int kind, const char* s1 )
{
	fprintf( out, "==Block\r%d\r%d\r%d\r%d\r0\r%s\r", id++, kind, x, y, s1 );
	fflush( out );
	x+=30; y+=40;
}


void block2( FILE* out, int kind, const char* s1, const char* s2 )
{
	fprintf( out, "==Block\r%d\r%d\r%d\r%d\r0\r%s\r%s\r", id++, kind, x, y, s1, s2 );
	fflush( out );
	x+=30; y+=40;
}


void block3( FILE* out, int kind, const char* s1, const char* s2, const char* s3 )
{
	fprintf( out, "==Block\r%d\r%d\r%d\r%d\r0\r%s\r%s\r%s\r", id++, kind, x, y, s1, s2, s3 );
	fflush( out );
	x+=30; y+=40;
}


void block4( FILE* out, int kind, const char* s1, const char* s2, const char* s3, const char* s4 )
{
	fprintf( out, "==Block\r%d\r%d\r%d\r%d\r0\r%s\r%s\r%s\r%s\r", id++, kind, x, y, s1, s2, s3, s4 );
	fflush( out );
	x+=30; y+=40;
}


void block6( FILE* out, int kind, const char* s1, const char* s2, const char* s3,
			const char* s4, const char* s5, const char* s6 )
{
	fprintf( out, "==Block\r%d\r%d\r%d\r%d\r0\r%s\r%s\r%s\r%s\r%s\r%s\r", id++, kind, x, y, s1, s2, s3, s4, s5, s6 );
	fflush( out );
	x+=30; y+=40;
}


/**
 * The condition part of an if.  Expr must be a comparative expression and
 * invert indicates that we want the NOT of the condition.
 */
void ifblock( FILE* out, Expression* expr, int invert )
{
	char left[256], right[256];
	fprintf( out, "==Block\r%d\r%d\r%d\r%d\r0\r", id++, IFTHENELSE, x, y );
	switch( expr->kind )
	{
	case equality_expr:
		if( !invert ) fprintf( out, "=\r" );
		else codegen_error( "Invalid comparative expression (==).", expr->lineNo );
		break;		
	case lessthan_expr:
		if( !invert ) fprintf( out, "<\r" );
		else fprintf( out, ">=\r" );
		break;
	case greaterthan_expr:
		if( !invert ) fprintf( out, ">\r" );
		else fprintf( out, "<=\r" );
		break;
	case lessequal_expr:
		if( !invert ) fprintf( out, "<=\r" );
		else fprintf( out, ">\r" );
		break;
	case greaterequal_expr:
		if( !invert ) fprintf( out, ">=\r" );
		else fprintf( out, "<\r" );
		break;
	case inequality_expr:
		if( invert ) fprintf( out, "=\r" );
		else codegen_error( "Invalid comparative expression (!=).", expr->lineNo );
		break;
	default:
		codegen_error( "Invalid comparative expression.", expr->lineNo );
		return;
	}

	fprintf( out, "%s\r%s\r",
		getSimpleExpression( left, expr->data.binary_op_part.left ),
		getSimpleExpression( right, expr->data.binary_op_part.right ) );
	fflush( out );
	x+=30; y+=40;	
}


/**
 * An assignment statement (corresponding to an arithmetic block).  name is the
 * variable to assign to (a-l) and expr is a simple binary expression or a
 * variable or integer on its own.
 */
void assignblock( FILE* out, Ident* name, Expression* expr )
{
	char buf[256], left[256], right[256];

	fprintf( out, "==Block\r%d\r%d\r%d\r%d\r0\r", id++, ARITHMETIC, x, y );
	x+=30; y+=40;

	if( (expr->kind == variable_expr) || (expr->kind == integer_expr) )
	{
		fprintf( out, "None\r%s\rc\r%s\r", getSimpleExpression( buf, expr ), name->name  );
		return;
	}

	switch( expr->kind )
	{
	case addition_expr:		  fprintf( out, "+\r" ); break;
	case subtraction_expr:	  fprintf( out, "-\r" ); break;
	case multiplication_expr: fprintf( out, "*\r" ); break;
	case division_expr:		  fprintf( out, "/\r" ); break;
	default:
		codegen_error( "Illegal expression in assignment.", expr->lineNo );
		break;
	}

	fprintf( out, "%s\r%s\r%s\r",
		getSimpleExpression( left, expr->data.binary_op_part.left ),
		getSimpleExpression( right, expr->data.binary_op_part.right ),
		name->name );
	fflush( out );
}

void link( FILE* out, int fid, int fport, int tid, int tport )
{
	fprintf( out, "==Link\r%d\r%d\r%d\r%d\r255\r", fid, fport, tid, tport );
	fflush( out );
}


/**
 * Produces links from each of the blocks/ports on the prev stack to the given
 * block/port.
 */
void joinTo( int tid, int tport )
{
	for( ; prevc > 0; prevc-- )
		lines = newLineList( newLine( previd[prevc-1], prevport[prevc-1], tid, tport ), lines );
}


Expression *findArgumentValue( IdentList *arglist, ExpressionList *parmlist, Ident *variable )
{
	for( ; arglist && parmlist; arglist = arglist->tail, parmlist = parmlist->tail )
	{
		if( strcmp( arglist->head->name, variable->name ) == 0 )
			return parmlist->head;
	}

	return NULL;
}


/**
 * Generate an call to an output or process.
 */
void generateClassCallStatement( FILE* out, Statement* stmt )
{
	Member *member = stmt->data.call_part.definition.member;

	if( !member )
	{
		codegen_error2( "Output or process member call %s.%s() not linked.",
			stmt->lineNo, stmt->data.call_part.clazz->name,
			stmt->data.call_part.name->name );
		return;
	}
	
	/* Have we been linked? */
	if( !member->data.normal_part.definition )
	{
		codegen_error2( "%s.%s() has no definition.", stmt->lineNo,
			stmt->data.call_part.clazz->name, stmt->data.call_part.name->name );
		return;
	}

	/* Make sure we have a block */
	if( member->data.normal_part.definition->kind == block_def )
	{
		ExpressionList *exprlist = member->data.normal_part.definition->pArgs;
		char buf[256];

		/* Is there at least one argument */
		if( exprlist )
		{
			fprintf( out, "==Block\r%d\r", id++ );

			/* First argument is block id */
			fprintf( out, "%s\r", getSimpleExpression( buf, exprlist->head ) );
			fprintf( out, "%d\r%d\r0\r", x, y );
			x+=30; y+=40;

			/* Output subsequent arguments */
			for( exprlist = exprlist->tail; exprlist; exprlist = exprlist->tail )
			{
				if( exprlist->head->kind == variable_expr )
				{
					Expression *parm = findArgumentValue(
						member->data.normal_part.pArgs,
						stmt->data.call_part.pArgs,
						exprlist->head->data.variable_part );

					if( !parm )
						parm = findArgumentValue(
							member->parent->pArgs,
							stmt->data.call_part.pArgs,
							exprlist->head->data.variable_part );

					if( parm )
						fprintf( out, "%s\r", getSimpleExpression( buf, parm ) );
					else
					{
						codegen_error1( "Could not resolve argument '%s'.",
							stmt->lineNo,
							exprlist->head->data.variable_part->name );
						break;
					}
				}
				else
					fprintf( out, "%s\r", getSimpleExpression( buf, exprlist->head ) );
			}

			fflush( out );
		}
		else
			codegen_error2( "Insufficent arguments to %s.%s()", stmt->lineNo,
				stmt->data.call_part.clazz->name, stmt->data.call_part.name->name );
	}
	else
		codegen_error2( "%s.%s() is not of block type.", stmt->lineNo,
			stmt->data.call_part.clazz->name, stmt->data.call_part.name->name );
}


/**
 * Generate a call to a input (used as part of a switch).
 */
void generateClassCallExpression( FILE* out, Expression* expr )
{
	Member *member = expr->data.call_part.definition;

	if( !member )
	{
		codegen_error2( "Input member call %s.%s() not linked.", expr->lineNo,
			expr->data.call_part.clazz->name, expr->data.call_part.name->name );
		return;
	}

	if( member->kind != input_mbr )
	{
		codegen_error2( "%s.%s() is not an input member.", expr->lineNo,
			expr->data.call_part.clazz->name, expr->data.call_part.name->name );
		return;
	}

	if( !member->data.input_part.definition )
	{
		codegen_error2( "%s.%s() has no definition.", expr->lineNo,
			expr->data.call_part.clazz->name, expr->data.call_part.name->name );
		return;
	}

	if( member->data.input_part.definition->kind == block_def )
	{
		ExpressionList *exprlist = member->data.input_part.definition->pArgs;
		char buf[256];

		if( exprlist )
		{
			fprintf( out, "==Block\r%d\r", id++ );
			fprintf( out, "%s\r", getSimpleExpression( buf, exprlist->head ) );
			fprintf( out, "%d\r%d\r0\r", x, y );
			x+=30; y+=40;

			for( exprlist = exprlist->tail; exprlist; exprlist = exprlist->tail )
			{
				if( exprlist->head->kind == variable_expr )
				{
					Expression *parm = findArgumentValue(
						member->data.input_part.pArgs,
						expr->data.call_part.pArgs,
						exprlist->head->data.variable_part );

					if( !parm )
						parm = findArgumentValue(
							member->parent->pArgs,
							expr->data.call_part.pArgs,
							exprlist->head->data.variable_part );

					if( parm )
						fprintf( out, "%s\r", getSimpleExpression( buf, parm ) );
					else
						codegen_error1( "Could not resolve argument to %s.",
							exprlist->head->lineNo,
							exprlist->head->data.variable_part->name );
				}
				else
					fprintf( out, "%s\r", getSimpleExpression( buf, exprlist->head ) );
			}

			fflush( out );
		}
		else
			codegen_error2( "Insufficent arguments to %s.%s()", expr->lineNo,
				expr->data.call_part.clazz->name, expr->data.call_part.name->name );
	}
	else
		codegen_error2( "%s.%s() is not of block type.", expr->lineNo,
			expr->data.call_part.clazz->name, expr->data.call_part.name->name );
}


/**
 * Called to generate a block for the given Statement.
 */
void generateStatement( FILE* out, Statement* stmt )
{
	switch( stmt->kind )
	{
	case if_stmt:
		if( stmt->data.if_part.condition )
		{
		/* Generate the if itself */
		int thenid[MAX_DEPTH], thenport[MAX_DEPTH], thenc = 0;
		int ifid = id, ifx = x, ify = y;
		int invert = (stmt->data.if_part.condition->kind == inequality_expr);
		joinTo( ifid, 1 );
		PUSH_PREV( ifid, (invert ? 2 : 1) );
		ifblock( out, stmt->data.while_and_do_part.condition, invert );
		/* Generate the if body */
		x += 60; y -= 80;
		generateStatementList( out, stmt->data.if_part.pThen );
		/* Save the end of the then */
		SAVE_PREV(then);
		/* Generate the optional else clause */
		PUSH_PREV( ifid, (invert ? 1 : 2) );
		x = ifx + 30; y = ify + 40;
		generateStatementList( out, stmt->data.if_part.pElse );
		/* Restore the end of the then onto the end of the else */
		RESTORE_PREV(then);
		}
		else
			codegen_error( "missing condition in if statement.", stmt->lineNo );
		break;
	case while_stmt:
		if( stmt->data.while_and_do_part.condition->kind == boolean_expr )
		{
			/* while( true ) [while( false ) optimises to nothing] */
			if( stmt->data.while_and_do_part.condition->data.integer_part == 1 )
			{
				char labbuf[256];
				int whilex = x, whiley = y;
				sprintf( labbuf, "%d", landid++ );
				/* Create a land point for the start of the loop */
				PUSH_PREV( id, 1 );
				x = whilex - 30; y = whiley + 40;
				block1( out, LAND, labbuf );
				/* Generate the loop body - this will hook both before code and land to itself */
				x = whilex + 60; y = whiley;
				generateStatementList( out, stmt->data.while_and_do_part.pBody );
				/* Generate a jump point and link it to the last statement */
				joinTo( id, 1 );
				block1( out, JUMP, labbuf );
				/* There is no way out so reset prev counter */
				prevc = 0;
			}
		}
		else
		{
			/* Ordinary while loop */
			char labbuf[256];
			int whileid, whilex = x, whiley = y;
			int invert = (stmt->data.while_and_do_part.condition->kind == inequality_expr);
			sprintf( labbuf, "%d", landid++ );
			/* Create a land point for the start of the loop */
			PUSH_PREV( id, 1 );
			x = whilex - 90; y = whiley + 40;
			block1( out, LAND, labbuf );
			/* Generate the loop condition - this will hook both before code and land to itself */
			whileid = id; x = whilex; y = whiley;
			joinTo( whileid, 1 );
			PUSH_PREV( whileid, (invert ? 2 : 1) );
			ifblock( out, stmt->data.while_and_do_part.condition, invert );
			/* Generate the loop body - this will hook to the condition */
			x += 60; y -= 80;
			generateStatementList( out, stmt->data.while_and_do_part.pBody );
			/* Generate a jump point and link it to the last statement */
			joinTo( id, 1 );
			block1( out, JUMP, labbuf );
			/* Remember the other branch of the condition */
			PUSH_PREV( whileid, (invert ? 1 : 2) );
			x = whilex + 30; y = whiley + 40;
		}
		break;
	case do_stmt:
		if( stmt->data.while_and_do_part.condition->kind == boolean_expr )
		{
			if( stmt->data.while_and_do_part.condition->data.integer_part == 1 )
			{
				/* do {} while( true ) */
				char labbuf[256];
				int whilex = x, whiley = y;
				sprintf( labbuf, "%d", landid++ );
				/* Create a land point for the start of the loop */
				PUSH_PREV( id, 1 );
				x = whilex - 30; y = whiley + 40;
				block1( out, LAND, labbuf );
				/* Generate the loop body - this will hook both before code and land to itself */
				x = whilex + 60; y = whiley;
				generateStatementList( out, stmt->data.while_and_do_part.pBody );
				/* Generate a jump point and link it to the last statement */
				joinTo( id, 1 );
				block1( out, JUMP, labbuf );
				/* There is no way out so reset prev counter */
				prevc = 0;
			}
			else
			{
				/* do {} while( false ) - optimises to single iteration of body */
				generateStatementList( out, stmt->data.while_and_do_part.pBody );
			}
		}
		else
		{
			/* Ordinary do {} while loop */
			char labbuf[256];
			int whileid, whilex, whiley;
			int dox = x, doy = y;
			int invert = (stmt->data.while_and_do_part.condition->kind == inequality_expr);
			sprintf( labbuf, "%d", landid++ );
			/* Create a land point for the start of the loop */
			PUSH_PREV( id, 1 );
			x = dox - 90; y = doy + 40;
			block1( out, LAND, labbuf );
			/* Generate the loop body - this will hook to the condition and land to itself */
			x = dox; y = doy;
			generateStatementList( out, stmt->data.while_and_do_part.pBody );
			/* Generate the loop condition - this will hook to the end of the body */
			whileid = id; whilex = x; whiley = y;
			joinTo( whileid, 1 );
			PUSH_PREV( whileid, (invert ? 2 : 1) );
			ifblock( out, stmt->data.while_and_do_part.condition, invert );
			/* Generate a jump point and link it to the while statement */
			x += 60; y -= 80;
			joinTo( id, 1 );
			block1( out, JUMP, labbuf );
			/* Remember the other branch of the condition */
			PUSH_PREV( whileid, (invert ? 1 : 2) );
			x = whilex + 30; y = whiley + 40;
		}
		break;
	case switch_stmt:
		if( stmt->data.switch_part.expr
			&& (stmt->data.switch_part.expr->kind == call_expr)
			&& stmt->data.switch_part.expr->data.call_part.clazz )
		{
			CaseBlockList *cblist;
			int breakid[MAX_DEPTH], breakport[MAX_DEPTH], breakc = 0;
			int switchid = id, switchx = x, switchy = y;
			joinTo( id, 1 );
			generateClassCallExpression( out, stmt->data.switch_part.expr );
			for( cblist = stmt->data.switch_part.pCaseBlocks; cblist; cblist = cblist->tail )
			{
				ExpressionList *exprlist, *defaultlist;
				StatementList *stmtlist;
				x = switchx + 90; y = switchy - 160; switchy += 80;
				for( exprlist = cblist->head->pCases; exprlist; exprlist = exprlist->tail )
				{
					/* Remember each of the ports on the switch block to link from */
					switch( exprlist->head->kind )
					{
					case integer_expr:
						PUSH_PREV( switchid, exprlist->head->data.integer_part );
						break;
					case constant_expr:
						if( exprlist->head->data.constant_part.value
							&& (exprlist->head->data.constant_part.value->kind == integer_expr) )
						{
							PUSH_PREV( switchid, exprlist->head->data.constant_part.value->data.integer_part );
						}
						else
							codegen_error( "case with non-integer value.", exprlist->head->lineNo );
						break;
					case default_expr:
						for( defaultlist = exprlist->head->data.default_part.pCases; defaultlist; defaultlist = defaultlist->tail )
						{
							PUSH_PREV( switchid, defaultlist->head->data.integer_part );
						}
						break;
					default:
						codegen_error( "case with non-constant value.", exprlist->head->lineNo );
						break;
					}
				}
				for( stmtlist = cblist->head->pStmts; stmtlist; stmtlist = stmtlist->tail )
				{
					/* If we encounter a break remember the last statement(s) */
					if( stmtlist->head->kind == break_stmt )
					{
						SAVE_PREV(break);
					}
					else /* otherwise output as normal */
						generateStatement( out, stmtlist->head );
				}
			}
			/* Take each statement before the breaks and get it ready to make a link */
			RESTORE_PREV(break);
			x += 60;
		}
		else
			codegen_error( "switch can only be performed on a sensor value.", stmt->lineNo );
		break;
	case assignment_stmt:
		joinTo( id, 1 );
		PUSH_PREV( id, 1 );
		if( stmt->data.assignment_part.kind == equals_assign )
			assignblock( out, stmt->data.assignment_part.name, stmt->data.assignment_part.value );
		else
			codegen_error( "Can only generate code for = assignments.", stmt->lineNo );
		break;
	case call_stmt:
		if( stmt->data.call_part.clazz != NULL )
		{
			joinTo( id, 1 );
			if( strcmp( stmt->data.call_part.name->name, "stop" ) != 0 )
			{
				PUSH_PREV( id, 1 );
			}
			generateClassCallStatement( out, stmt );
		}
		else
			codegen_error( "Procedure calls not supported yet.", stmt->lineNo );
		break;
	case return_stmt:
		codegen_error( "return not yet implemented.", stmt->lineNo );
		break;
	case break_stmt:
		codegen_error( "break (from loop) not yet implemented.", stmt->lineNo );
		break;
	case start_stmt:
		/* no line to draw here */
		PUSH_PREV( id, 1 );
		block2( out, START, "0", "[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "
				"0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "
				"0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "
				"0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "
				"0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "
				"0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "
				"0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "
				"0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "
				"0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "
				"0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "
				"0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "
				"0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "
				"0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "
				"0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "
				"0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "
				"0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]" );
		break;
	case goto_stmt:
		codegen_error( "goto not yet implemented.", stmt->lineNo );
		break;
	case label_stmt:
		codegen_error( "label not yet implemented.", stmt->lineNo );
		break;
	default:
	case for_stmt:	/* Should be transformed into a while statement */
		codegen_error( "Unexpected statement kind.", stmt->lineNo );
		break;
	}
}


/**
 * Iterates through a list of Statements.
 */
void generateStatementList( FILE* out, StatementList* stmtlist )
{
	for( ; stmtlist; stmtlist = stmtlist->tail )
		generateStatement( out, stmtlist->head );
}


/**
 * Generates the blocks for the given Procedure.
 */
void generateProcedure( FILE* out, Procedure* proc )
{
	prevc = 0;

	if( !proc->pStmts )
		return;

	generateStatementList( out, proc->pStmts );

	/* If there are open outputs, then insert a stop automatically */
	if( prevc > 0 )
	{
		joinTo( id, 1 );
		block0( out, STOP );
	}
}


/**
 * Called by the main program to generate all the code.  Note that only
 * procedures (subroutines) generate code.  Classes and instances are internal
 * structures and are simple used to define the inputs and outputs of the
 * robot.
 */
void generateCode( Program* prog )
{
	ProcedureList* proclist;
	FILE* out = fopen( prog->outputfile, "w" );

	if( !out )
	{
		codegen_error1( "unable to open output file '%s'.", NULL, prog->outputfile );
		return;
	}

	/* Reset the id */
	id = 401;
	landid = 1;
	x = 50; y = 30;

	fprintf( out, "==Complete\r" );
	fprintf( out, "1\r" );
	fflush( out );

	for( proclist = prog->pProcs; proclist; proclist = proclist->tail )
	{
		if( proclist->head )
		{
			if( strcmp( proclist->head->name->name, "main" ) == 0 )
				generateProcedure( out, proclist->head );
			else
				codegen_warning1( "Skipping procedure \"%s\".", NULL, proclist->head->name->name );
		}
	}

	prog->pLines = lines;

	for( ; lines; lines = lines->tail )
	{
		link( out, lines->head->fromid, lines->head->fromport,
			  lines->head->toid, lines->head->toport );
	}

	fprintf( out, "EoF" );
	fflush( out );

	fclose( out );
}
