/******************************************************************************
 *
 * Filename: dumpgen.c
 *
 * Author:   Robert Thornburrow
 * Project:  Cycle - The CYbot Control LanguagE
 * Created:  5th June 2003
 *
 * Purpose:  Parse tree de-compiler.
 *
 * (c) Copyright R.H.Thornburrow, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>

#include "cycdefs.h"


void dumpStatement( Statement* stmt, int indent );


char indentbuf[256];

const char* indentstr( int indent )
{
	char* buf = indentbuf;

	for( ; indent > 0; indent-- )
		buf += sprintf( buf, "    " );

	return indentbuf;
}


void dumpExpression( Expression* expr )
{
	if( !expr )
	{
		printf( "/* null */" );
		return;
	}

	switch( expr->kind )
	{
	case variable_expr:
		printf( expr->data.variable_part->name );
		break;
	case integer_expr:
		printf( "%d", expr->data.integer_part );
		break;
	case string_expr:
		printf( "\"%s\"", expr->data.string_part );
		break;
	case boolean_expr:
		printf( "%s", expr->data.integer_part ? "true" : "false" );
		break;
	case addition_expr:
		dumpExpression( expr->data.binary_op_part.left );
		printf( " + " );
		dumpExpression( expr->data.binary_op_part.right );
		break;
	case subtraction_expr:
		dumpExpression( expr->data.binary_op_part.left );
		printf( " - " );
		dumpExpression( expr->data.binary_op_part.right );
		break;
	case multiplication_expr:
		dumpExpression( expr->data.binary_op_part.left );
		printf( " * " );
		dumpExpression( expr->data.binary_op_part.right );
		break;
	case division_expr:
		dumpExpression( expr->data.binary_op_part.left );
		printf( " / " );
		dumpExpression( expr->data.binary_op_part.right );
		break;
	case equality_expr:
		dumpExpression( expr->data.binary_op_part.left );
		printf( " == " );
		dumpExpression( expr->data.binary_op_part.right );
		break;
	case inequality_expr:
		dumpExpression( expr->data.binary_op_part.left );
		printf( " != " );
		dumpExpression( expr->data.binary_op_part.right );
		break;
	case lessthan_expr:
		dumpExpression( expr->data.binary_op_part.left );
		printf( " > " );
		dumpExpression( expr->data.binary_op_part.right );
		break;
	case greaterthan_expr:
		dumpExpression( expr->data.binary_op_part.left );
		printf( " > " );
		dumpExpression( expr->data.binary_op_part.right );
		break;
	case lessequal_expr:
		dumpExpression( expr->data.binary_op_part.left );
		printf( " <= " );
		dumpExpression( expr->data.binary_op_part.right );
		break;
	case greaterequal_expr:
		dumpExpression( expr->data.binary_op_part.left );
		printf( " >= " );
		dumpExpression( expr->data.binary_op_part.right );
		break;
	case call_expr:
		if( expr->data.call_part.clazz )
			printf( "%s.%s()", expr->data.call_part.clazz->name, expr->data.call_part.name->name );
		else
			printf( "%s()", expr->data.call_part.name->name );
		break;
	case constant_expr:
		printf( "%s.%s", expr->data.constant_part.clazz->name, expr->data.constant_part.name->name );
		break;
	case default_expr:
		printf( "/* default */" );
		break;
	default:
		printf( "/* unknown */" );
		break;
	}
}


void dumpStatementBlock( StatementList* stmtlist, int indent )
{
	printf( "%s{\n", indentstr(indent) );

	for( ; stmtlist; stmtlist = stmtlist->tail )
		dumpStatement( stmtlist->head, indent+1 );

	printf( "%s}\n", indentstr(indent) );
}


void dumpAssignmentStatement( Statement* stmt )
{
	printf( "%s", stmt->data.assignment_part.name->name );

	switch( stmt->data.assignment_part.kind )
	{
	case equals_assign:
		printf( " = " );
		dumpExpression( stmt->data.assignment_part.value );
		break;
	case plusequals_assign:
		printf( " += " );
		dumpExpression( stmt->data.assignment_part.value );
		break;
	case minusequals_assign:
		printf( " -= " );
		dumpExpression( stmt->data.assignment_part.value );
		break;
	case timesequals_assign:
		printf( " *= " );
		dumpExpression( stmt->data.assignment_part.value );
		break;
	case divideequals_assign:
		printf( " /= " );
		dumpExpression( stmt->data.assignment_part.value );
		break;
	case increment_assign:
		printf( "++" );
		break;
	case decrement_assign:
		printf( "--" );
		break;
	default:
		printf( " = /* unknown */" );
		break;
	}
}


void dumpStatement( Statement* stmt, int indent )
{
	if( !stmt )
	{
		printf( "%s/* null */\n", indentstr(indent) );
		return;
	}

	switch( stmt->kind )
	{
	case if_stmt:
		printf( "%sif( ", indentstr(indent) );
		dumpExpression( stmt->data.if_part.condition );
		if( stmt->data.if_part.pThen )
		{
			printf( " )\n" );
			dumpStatementBlock( stmt->data.if_part.pThen, indent );
		}
		else
			printf( " ) {}\n" );
		if( stmt->data.if_part.pElse )
		{
			printf( "%selse\n", indentstr(indent) );
			dumpStatementBlock( stmt->data.if_part.pElse, indent );
		}
		break;
	case while_stmt:
		printf( "%swhile( ", indentstr(indent) );
		dumpExpression( stmt->data.while_and_do_part.condition );
		if( stmt->data.while_and_do_part.pBody )
		{
			printf( " )\n" );
			dumpStatementBlock( stmt->data.while_and_do_part.pBody, indent );
		}
		else
			printf( " ) {}\n" );
		break;
	case do_stmt:
		printf( "%sdo", indentstr(indent) );
		dumpStatementBlock( stmt->data.while_and_do_part.pBody, indent );
		printf( "%swhile( ", indentstr(indent) );
		dumpExpression( stmt->data.while_and_do_part.condition );
		printf( " )\n" );
		break;
	case for_stmt:
		printf( "%sfor( ", indentstr(indent) );
		dumpAssignmentStatement( stmt->data.for_part.initial );
		printf( "; " );
		dumpExpression( stmt->data.for_part.condition );
		printf( "; " );
		dumpAssignmentStatement( stmt->data.for_part.increment );
		if( stmt->data.for_part.pBody )
		{
			printf( " )\n" );
			dumpStatementBlock( stmt->data.for_part.pBody, indent );
		}
		else
			printf( " ) {}\n" );
		break;
	case switch_stmt:
		printf( "%sswitch( ", indentstr(indent) );
		dumpExpression( stmt->data.switch_part.expr );
		if( stmt->data.switch_part.pCaseBlocks )
		{
			CaseBlockList *cblist;
			printf( " )\n" );
			printf( "%s{\n", indentstr(indent) );
			for( cblist = stmt->data.switch_part.pCaseBlocks; cblist; cblist = cblist->tail )
			{
				ExpressionList *exprlist;
				StatementList *stmtlist;
				for( exprlist = cblist->head->pCases; exprlist; exprlist = exprlist->tail )
				{
					if( exprlist->head && (exprlist->head->kind == default_expr) )
						printf( "%sdefault", indentstr(indent) );
					else
					{
						printf( "%scase ", indentstr(indent) );
						dumpExpression( exprlist->head );
					}
					printf( ":\n" );
				}
				for( stmtlist = cblist->head->pStmts; stmtlist; stmtlist = stmtlist->tail )
					dumpStatement( stmtlist->head, indent+1 );				
			}
			printf( "%s}\n", indentstr(indent) );
		}
		else
			printf( " ) {}\n" );
		break;
	case assignment_stmt:
		printf( "%s", indentstr(indent) );
		dumpAssignmentStatement( stmt );
		printf( ";\n" );
		break;
	case call_stmt:
		if( stmt->data.call_part.clazz )
			printf( "%s%s.%s(", indentstr(indent), stmt->data.call_part.clazz->name, stmt->data.call_part.name->name );
		else
			printf( "%s%s(", indentstr(indent), stmt->data.call_part.name->name );
		if( stmt->data.call_part.pArgs )
		{
			ExpressionList *exprlist;
			printf( " " );
			for( exprlist = stmt->data.call_part.pArgs; exprlist; exprlist = exprlist->tail )
			{
				dumpExpression( exprlist->head );
				if( exprlist->tail )
					printf( ", " );
			}
			printf( " " );
		}
		printf( ");\n" );
		break;
	case return_stmt:
		printf( "%sreturn;\n", indentstr(indent) );
		break;
	case break_stmt:
		printf( "%sbreak;\n", indentstr(indent) );
		break;
	case start_stmt:
		printf( "%sstart;\n", indentstr(indent) );
		break;
	case goto_stmt:
		printf( "%sgoto ;\n", indentstr(indent) );
		break;
	case label_stmt:
		printf( "%s: \n", indentstr(indent) );
		break;
	default:
		printf( "%s/* unknown statement */\n", indentstr(indent) );
		break;
	}
}


void dumpProcedure( Procedure* proc )
{
	StatementList* stmtlist;

	if( !proc )
	{
		printf( "/* null procedure */\n" );
		return;
	}

	/* Name of the procedure */
	printf( "proc %s(", proc->name->name );
	if( proc->pArgs )
	{
		IdentList* arglist;
		printf( " " );
		for( arglist = proc->pArgs; arglist; arglist = arglist->tail )
		{
			printf( "%s", arglist->head->name );
			if( arglist->tail)
				printf( ", " );
		}
		printf( " " );
	}
	printf( ")" );

	/* Body */
	printf( "\n{\n" );

	for( stmtlist = proc->pStmts; stmtlist; stmtlist = stmtlist->tail )
	{
		if( stmtlist->head )
			dumpStatement( stmtlist->head, 1 );
	}

	printf( "}\n" );
}


void dumpInstance( Instance* inst )
{
	if( !inst )
	{
		printf( "/* null instance */\n" );
		return;
	}

	printf( "%s %s", inst->type->name, inst->name->name );

	/* Arguments */
	if( inst->pArgs )
	{
		ExpressionList *exprlist;
		printf( "( " );
		for( exprlist = inst->pArgs; exprlist; exprlist = exprlist->tail )
		{
			if( exprlist->head )
				dumpExpression( exprlist->head );
			if( exprlist->tail )
				printf( ", " );
		}
		printf( " )" );
	}

	printf( ";\n" );
}


void dumpMemberDef( MemberDef* def )
{
	switch( def->kind )
	{
	case block_def:
		printf( "block(" );
		if( def->pArgs )
		{
			ExpressionList *exprlist;
			printf( " " );
			for( exprlist = def->pArgs; exprlist; exprlist = exprlist->tail )
			{
				dumpExpression( exprlist->head );
				if( exprlist->tail)
					printf( ", " );
			}
			printf( " " );
		}
		printf( ");" );
		break;
	}
}


void dumpMember( Member* mbr, int indent )
{
	IdentList* arglist;

	if( mbr->kind == const_mbr )
	{
		printf( "%sconst %s = ", indentstr(indent), mbr->name->name );
		dumpExpression( mbr->data.const_part.value );
		printf( ";\n" );
		return;
	}

	printf( "%s", indentstr(indent) );

	switch( mbr->kind )
	{
	case input_mbr:	  arglist = mbr->data.input_part.pArgs; printf( "input" ); break;
	case process_mbr: arglist = mbr->data.normal_part.pArgs; printf( "process" ); break;
	case output_mbr:  arglist = mbr->data.normal_part.pArgs; printf( "output" ); break;
	}

	printf( " %s(", mbr->name->name );
	if( mbr->data.normal_part.pArgs )
	{
		printf( " " );
		for( ; arglist; arglist = arglist->tail )
		{
			printf( "%s", arglist->head->name );
			if( arglist->tail)
				printf( ", " );
		}
		printf( " " );
	}
	printf( ")" );

	/* Definition */
	if( !mbr->data.normal_part.definition )
	{
		printf( ";\n" );
		return;
	}

	printf( " : " );

	dumpMemberDef( mbr->data.normal_part.definition );

	printf( "\n" );
}


void dumpClass( Class* clazz )
{
	MemberList *mbrlist;

	if( !clazz )
	{
		printf( "/* null class */\n" );
		return;
	}

	/* Name of the class */
	printf( "class %s", clazz->name->name );

	/* Arguments */
	if( clazz->pArgs )
	{
		IdentList *idlist;
		printf( "( " );
		for( idlist = clazz->pArgs; idlist; idlist = idlist->tail )
		{
			if( idlist->head )
				printf( "%s", idlist->head->name );
			if( idlist->tail )
				printf( ", " );
		}
		printf( " )" );
	}

	/* Body */
	if( !clazz->pMembers )
	{
		printf( ";\n" );
		return;
	}

	printf( "\n{\n" );

	for( mbrlist = clazz->pMembers; mbrlist; mbrlist = mbrlist->tail )
	{
		if( mbrlist->head )
			dumpMember( mbrlist->head, 1 );
	}

	printf( "}\n" );
}


void dumpCode( Program* prog )
{
	PICList *piclist;

	for( piclist = prog->pPICs; piclist; piclist = piclist->tail )
	{
		switch( piclist->head->kind )
		{
		case class_pic:
			dumpClass( piclist->head->data.clazz );
			break;
		case instance_pic:
			dumpInstance( piclist->head->data.instance );
			break;
		case proc_pic:
			dumpProcedure( piclist->head->data.proc );
			break;
		}

		if( piclist->tail )
			printf( "\n" );
	}
}
