/******************************************************************************
 *
 * Filename: semantic.c
 *
 * Author:   Robert Thornburrow
 * Project:  Cycle - The CYbot Control LanguagE
 * Created:  7th June 2003
 *
 * Purpose:  Semantic parser.
 *
 * (c) Copyright R.H.Thornburrow, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cycdefs.h"
#include "errors.h"
#include "hashtable.h"


void checkStatementList( StatementList *stmtlist );
void checkStatement( Statement *stmt );
void checkConstantExpression( Expression *expr, int forCase, int forAssign );


Ident *findMatchingArgument( IdentList *arglist, Ident *name )
{
	for( ; arglist; arglist = arglist->tail )
	{
		if( strcmp( name->name, arglist->head->name ) == 0 )
			return arglist->head;
	}

	return NULL;
}


int getIntegerValue( Expression *expr )
{
	switch( expr->kind )
	{
	case integer_expr:
		return expr->data.integer_part;
	case constant_expr:
		return getIntegerValue( expr->data.constant_part.value );
	default:
		return 0;
	}
}


void checkMemberArgumentList( Hashtable *classArgsHashtable, IdentList *pArgs )
{
	Hashtable* memberArgsHashtable = Hashtable_new( 16 );

	for( ; pArgs; pArgs = pArgs->tail )
	{
		if( Hashtable_get( memberArgsHashtable, pArgs->head->name ) ) {
			semantic_error1( "argument name '%s' already in use.", pArgs->head->lineNo, pArgs->head->name );
		} else {
			if( Hashtable_get( classArgsHashtable, pArgs->head->name ) )
				semantic_warning1( "argument name '%s' overrides class argument.", pArgs->head->lineNo, pArgs->head->name );

			Hashtable_put( memberArgsHashtable, pArgs->head->name, pArgs->head );
		}
	}

	Hashtable_delete( memberArgsHashtable, 0, 0, 0, NULL );
}


void checkParmList( Member *member, Class *clazz, IdentList *pMemberArgs, ExpressionList *parmlist )
{
	if( !parmlist )
	{
		semantic_error2( "definition of %s.%s() has too few arguments.",
				member->name->lineNo,
				member->parent->name->name,
				member->name->name );
		return;
	}

	/* For each parameter in the definition: */
	for( ; parmlist; parmlist = parmlist->tail )
	{
		if( parmlist->head )
		{
			switch( parmlist->head->kind )
			{
			case variable_expr:
				/* Variable: find matching argument name, first in member declaration, then in class declaration */
				if( !findMatchingArgument( pMemberArgs, parmlist->head->data.variable_part )
					&& !findMatchingArgument( clazz->pArgs, parmlist->head->data.variable_part ) )
					semantic_error1( "'%s' not declared.",
						parmlist->head->data.variable_part->lineNo,
						parmlist->head->data.variable_part->name );
				break;
			case integer_expr:
			case string_expr:
				/* Integer or String constant: no need to do anything */
				break;
			case constant_expr:
				checkConstantExpression( parmlist->head, FALSE, FALSE );
				break;
			default:
				/* Anything else is an error */
				semantic_error1( "'%s' is not an integer, string or variable name.",
						parmlist->head->data.variable_part->lineNo,
						parmlist->head->data.variable_part->name );
				break;
			}
		}
	}
}


void checkMember( Class *clazz, Hashtable *classArgsHashtable, Member *member )
{
	/* Fill in the parent Class of this member */
	member->parent = clazz;

	switch( member->kind )
	{
	case const_mbr:
		/* All constants must be integers */
		if( !member->data.const_part.value )
		{
			semantic_error2( "%s.%s has no value.",
					member->name->lineNo,
					member->parent->name->name,
					member->name->name );
			return;
		}

		switch( member->data.const_part.value->kind )
		{
		case integer_expr:
		case string_expr:
			break;
		default:
			semantic_error( "constant must be an integer or string value.", member->name->lineNo );
			break;
		}
		break;
	case input_mbr:
		/* Check we have a definition */
		if( !member->data.input_part.definition )
		{
			semantic_error2( "%s.%s() has no definition.",
					member->name->lineNo,
					member->parent->name->name,
					member->name->name );
			return;
		}

		/* Check we have a valid range (if one was specified) */
		if( member->data.input_part.range )
		{
			Range * range = member->data.input_part.range;

			if( (range->from->kind != integer_expr) || (range->to->kind != integer_expr) )
			{
				semantic_error( "range of input member must be specified using integers only.",
					range->from->lineNo );
				return;
			}

			if( range->from->data.integer_part > range->to->data.integer_part )
			{
				semantic_error( "min must be less than max for range of input member.",
					range->from->lineNo );
				return;
			}
		}

		checkMemberArgumentList( classArgsHashtable, member->data.input_part.pArgs );
		checkParmList( member, clazz, member->data.input_part.pArgs, member->data.input_part.definition->pArgs );
		break;
	case output_mbr:
	case process_mbr:
		/* Check we have a definition */
		if( !member->data.normal_part.definition )
		{
			semantic_error2( "%s.%s() has no definition.",
					member->name->lineNo,
					member->parent->name->name,
					member->name->name );
			return;
		}

		checkMemberArgumentList( classArgsHashtable, member->data.normal_part.pArgs );
		checkParmList( member, clazz, member->data.normal_part.pArgs, member->data.normal_part.definition->pArgs );
		break;
	}
}


void checkClass( Class *clazz )
{
	Hashtable *classArgsHashtable = Hashtable_new( 16 );
	IdentList *arglist;
	MemberList *mbrlist;

	for( arglist = clazz->pArgs; arglist; arglist = arglist->tail )
	{
		if( Hashtable_get( classArgsHashtable, arglist->head->name ) )
			semantic_error1( "argument name '%s' already in use.", arglist->head->lineNo, arglist->head->name );
		else
			Hashtable_put( classArgsHashtable, arglist->head->name, arglist->head );
	}


	for( mbrlist = clazz->pMembers; mbrlist; mbrlist = mbrlist->tail )
	{
		if( mbrlist->head )
			checkMember( clazz, classArgsHashtable, mbrlist->head );
	}

	Hashtable_delete( classArgsHashtable, 0, 0, 0, NULL );
}

/********************************** Instances ********************************/

void checkInstance( Instance *instance )
{
	ExpressionList *parmlist;
	IdentList *arglist;

	/* Check that clazz is set */
	if( !instance->clazz )
		semantic_error2( "class %s for instance %s not found.",
			instance->name->lineNo, instance->type->name, instance->name->name );

	/* For each parameter in the instantiation: */
	for( arglist = instance->clazz->pArgs, parmlist = instance->pArgs;
		arglist && parmlist; arglist = arglist->tail, parmlist = parmlist->tail )
	{
		switch( parmlist->head->kind )
		{
		case integer_expr:
		case string_expr:
			/* Integer or String constant: no need to do anything */
			break;
		case constant_expr:
			checkConstantExpression( parmlist->head, FALSE, FALSE );
			break;
		default:
			/* Anything else is an error */
			semantic_error1( "%s is not an integer, string or constant.",
					parmlist->head->data.variable_part->lineNo,
					parmlist->head->data.variable_part->name );
			break;
		}
	}

	/* Must be no arguments or parameters left over */
	if( arglist || parmlist ) {
		semantic_error1( "incorrect number of arguments to class %s.",
			instance->name->lineNo, instance->type->name );
		return;
	}
}

/********************************** Procedures *******************************/

void checkVariableName( Ident* var )
{
	/* Variable name must be a letter between 'a' and 'l' */
	if( (strlen( var->name ) != 1) || (var->name[0] < 'a') || (var->name[0] > 'l') )
		semantic_error1( "'%s' is not a valid variable name.", var->lineNo, var->name );
}


void checkConstantExpression( Expression *expr, int forCase, int forAssign )
{
	/* Must be qualified by a class name */
	if( !expr->data.constant_part.clazz ) {
		semantic_error( "constants must be qualified by a class name.", expr->data.constant_part.name->lineNo );
		return;
	}

	/* Constant must exist and be linked */
	if( !expr->data.constant_part.value ) {
		semantic_error2( "%s.%s is not a constant.",
			expr->data.constant_part.name->lineNo,
			expr->data.constant_part.clazz->name,
			expr->data.constant_part.name->name );
		return;
	}

	/* If this is for a case statement, then must be an integer constant */
	if( forCase && (expr->data.constant_part.value->kind != integer_expr) )
	{
		semantic_error( "case values must be integers or integer constants.", expr->data.constant_part.name->lineNo );
		return;
	}

	/* If this is for a case statement, then must be an integer constant */
	if( forAssign && (expr->data.constant_part.value->kind != integer_expr) )
	{
		semantic_error( "Assignments may only use integers or integer constants.", expr->data.constant_part.name->lineNo );
		return;
	}
}


void checkConditionalExpression( Expression *expr, int noBoolean )
{
	if( !expr )
	{
		semantic_error( "missing condition expression.", NULL );
		return;
	}

	switch( expr->kind )
	{
	case boolean_expr:
		/* Check that boolean is in range */
		if( noBoolean )
			semantic_error( "Boolean constants not allowed here.", expr->lineNo );
		else if( (expr->data.integer_part != 0) && (expr->data.integer_part != 1) )
			semantic_error( "Boolean must be true or false.", expr->lineNo );
		break;
	case equality_expr:
	case inequality_expr:
	case lessthan_expr:
	case greaterthan_expr:
	case lessequal_expr:
	case greaterequal_expr:
		/* Validate left hand side is a variable or an integer */
		switch( expr->data.binary_op_part.left->kind )
		{
		case variable_expr:
			/* Variable name must be a letter between 'a' and 'l' */
			checkVariableName( expr->data.binary_op_part.left->data.variable_part );
			break;
		case integer_expr:
			/* Nothing to check for integers */
			break;
		default:
			/* Anything else is an error */
			semantic_error( "condition expressions must consist of only variables and integers.",
				expr->data.binary_op_part.left->lineNo );
			break;
		}
		/* Validate right hand side is a variable or an integer */
		switch( expr->data.binary_op_part.right->kind )
		{
		case variable_expr:
			/* Variable name must be a letter between 'a' and 'l' */
			checkVariableName( expr->data.binary_op_part.right->data.variable_part );
			break;
		case integer_expr:
			/* Nothing to check for integers */
			break;
		default:
			/* Anything else is an error */
			semantic_error( "condition expressions must consist of only variables and integers.",
				expr->data.binary_op_part.right->lineNo );
			break;
		}
		break;
	default:
		semantic_error( "invalid expression for condition.", expr->lineNo );
		break;
	}
}


void checkAssignmentExpression( AssignmentType assignkind, Expression *expr )
{
	switch( expr->kind )
	{
	case variable_expr:
		/* Variable name must be a letter between 'a' and 'l' */
		checkVariableName( expr->data.variable_part );
		break;
	case integer_expr:
		/* Nothing to check for integers */
		break;
	case addition_expr:
	case subtraction_expr:
	case multiplication_expr:
	case division_expr:
		/* Validate left hand side is a variable or an integer */
		switch( expr->data.binary_op_part.left->kind )
		{
		case variable_expr:
			/* Variable name must be a letter between 'a' and 'l' */
			checkVariableName( expr->data.binary_op_part.left->data.variable_part );
			break;
		case integer_expr:
			/* Nothing to check for integers */
			break;
		case constant_expr:
			checkConstantExpression( expr->data.binary_op_part.left, FALSE, TRUE );
			break;
		default:
			/* Anything else is an error */
			semantic_error( "assignment expressions must consist of only variables and integers.",
				expr->data.binary_op_part.left->lineNo );
			break;
		}
		/* Validate right hand side is a variable or an integer */
		switch( expr->data.binary_op_part.right->kind )
		{
		case variable_expr:
			/* Variable name must be a letter between 'a' and 'l' */
			checkVariableName( expr->data.binary_op_part.right->data.variable_part );
			break;
		case integer_expr:
			/* Nothing to check for integers */
			break;
		case constant_expr:
			checkConstantExpression( expr->data.binary_op_part.right, FALSE, TRUE );
			break;
		default:
			/* Anything else is an error */
			semantic_error( "assignment expressions must consist of only variables and integers.",
				expr->data.binary_op_part.right->lineNo );
			break;
		}
		break;
	case constant_expr:
		checkConstantExpression( expr, FALSE, TRUE );
		break;
	default:
		if( assignkind == equals_assign )
			semantic_error( "right-hand side must be a variable, integer or +, -, *, / expression.", expr->lineNo );
		else
			semantic_error( "right-hand side must be a variable, integer.", expr->lineNo );
		break;
	}
}


void checkSwitchExpression( Expression *expr )
{
	Member *member;
	IdentList *arglist;
	ExpressionList *parmlist;

	if( !expr )
	{
		semantic_error( "missing expression on switch.", NULL );
		return;
	}

	/* Make sure that the expression is a call */
	if( expr->kind != call_expr ) {
		semantic_error( "switch must be performed on an input member.", expr->lineNo );
		return;
	}

	/* Make sure that the expression is a member call */
	if( !expr->data.call_part.clazz ) {
		semantic_error( "switch must be performed on an input member.", expr->lineNo );
		return;
	}

	member = expr->data.call_part.definition;

	/* Make sure the called member is an input member which exists (has been linked) */
	if( !member || (member->kind != input_mbr) ) {
		semantic_error2( "%s.%s is not an input member.",
			expr->data.call_part.name->lineNo,
			expr->data.call_part.clazz->name,
			expr->data.call_part.name->name  );
		return;
	}

	for( arglist = member->data.input_part.pArgs, parmlist = expr->data.call_part.pArgs;
		arglist && parmlist; arglist = arglist->tail, parmlist = parmlist->tail )
	{
		switch( parmlist->head->kind )
		{
		case integer_expr:
		case string_expr:
			/* Nothing to check for integers */
			break;
		case constant_expr:
			checkConstantExpression( parmlist->head, FALSE, FALSE );
			break;
		default:
			/* Anything else is an error */
			semantic_error( "parameters must consist of only integers and constants.", parmlist->head->lineNo );
			break;
		}
	}

	/* Must be no arguments or parameters left over */
	if( arglist || parmlist ) {
		semantic_error2( "incorrect number of arguments to %s.%s().",
			expr->data.call_part.name->lineNo,
			expr->data.call_part.clazz->name,
			expr->data.call_part.name->name );
		return;
	}
}


void checkCaseExpression( Expression *expr )
{
	switch( expr->kind )
	{
	case integer_expr:
	case default_expr:
		/* Nothing to check for integers or default */
		break;
	case constant_expr:
		checkConstantExpression( expr, TRUE, FALSE );
		break;
	default:
		semantic_error( "case expression must be an integer or a constant", expr->lineNo );
		break;
	}
}


void checkCaseBlockList( CaseBlockList *cblklist )
{
	/* Loop through the case blocks */
	for( ; cblklist; cblklist = cblklist->tail )
	{
		ExpressionList *caselist;
		StatementList *stmtlist;
		int breakok = 0;

		for( caselist = cblklist->head->pCases; caselist; caselist = caselist->tail )
		{
			if( caselist->head )
				checkCaseExpression( caselist->head );
		}

		for( stmtlist = cblklist->head->pStmts; stmtlist; stmtlist = stmtlist->tail )
		{
			if( stmtlist->head )
			{
				if( stmtlist->head->kind == break_stmt )
				{
					breakok = 1;

					if( stmtlist->tail )
						semantic_error( "break must be the last statement in a case block.", stmtlist->head->lineNo );
				}
				else
					checkStatement( stmtlist->head );
			}
		}

		if( !breakok )
			semantic_error( "missing break at end of case block.",
				cblklist->head->pStmts ? cblklist->head->pStmts->head->lineNo : 0 );
	}
}


void checkSwitchCoverage( Statement *stmt )
{
	Range *range;
	int *tested, defaulted, from, to, length, okay, n, value;
	CaseBlockList *cblklist;
	Expression *defaultexpr;

	if( !stmt->data.switch_part.expr )
		return;

	/* Make sure this is a switch on an input */
	if( (stmt->data.switch_part.expr->kind != call_expr)
		|| !stmt->data.switch_part.expr->data.call_part.definition
		|| (stmt->data.switch_part.expr->data.call_part.definition->kind != input_mbr) )
		return;

	/* Get the range of the input */
	range = stmt->data.switch_part.expr->data.call_part.definition->data.input_part.range;

	/* Was a range specified on the input member? */
	if( !range )
		return;

	/* Get the actual range values */
	from = getIntegerValue( range->from );
	to   = getIntegerValue( range->to );

	/* Make sure from is less than (and not equal to) to */
	if( from >= to )
	{
		semantic_warning( "possible switch values have illegal range", stmt->lineNo );
		return;
	}

	/* Most switches begin a 1 (at least the ones I defined do) */
	if( from != 1 )
		semantic_warning( "possible switch values do not begin at 1", stmt->lineNo );

	/* Similarly, most switch ranges finish at 4 */
	if( to > 4 )
		semantic_warning( "more than 4 possible switch values", stmt->lineNo );

	/* If there is anything really daft, don't bother with the range check */
	if( (from < 1) || (to > 10) )
		return;

	/* Work out how many cases there are */
	length = to - from + 1;

	/* Flags for checking that all values have cases */
	defaulted = FALSE;
	tested = malloc( length * sizeof(int) );
	for( n = 0; n < length; n++ )  tested[n] = FALSE;

	/* Work through each of the case blocks */
	for( cblklist = stmt->data.switch_part.pCaseBlocks; cblklist; cblklist = cblklist->tail )
	{
		ExpressionList *caselist;

		/* For each case label: */
		for( caselist = cblklist->head->pCases; caselist; caselist = caselist->tail )
		{
			if( caselist->head )
			{
				/* If the switch contains a default, then don't report coverage */
				if( caselist->head->kind == default_expr )
				{
					if( defaulted )
					{
						semantic_error( "default may only be used once.", caselist->head->lineNo );
						break;
					}

					defaultexpr = caselist->head;
					defaulted = TRUE;
					continue;
				}

				/* Get the value of the case label */
				value = getIntegerValue( caselist->head );

				/* Check the value is in the desired range */
				if( (value < from) || (value > to) )
				{
					semantic_warning( "case value out of range", caselist->head->lineNo );
					continue;
				}

				/* Mark the value as tested */
				tested[value - from] = TRUE;
			}
		}
	}

	/* Check all the flags have been set */
	if( defaulted )
	{
		for( n = 0; n < length; n++ )
		{
			/* Make a list of each non-tested entry */
			if( !tested[n] )
				defaultexpr->data.default_part.pCases
					= newExpressionList(
						newIntegerExpr( integer_expr, from + n ),
							defaultexpr->data.default_part.pCases );
		}
	}
	else
	{
		okay = TRUE;

		/* Work through all the entries and if any are false, okay is false */
		for( n = 0; n < length; n++ )
		{
			if( !tested[n] )
			{
				okay = FALSE;
				break;
			}
		}

		/* Warn if not all flags were set */
		if( !okay )
			semantic_warning( "switch statement does not cover all possible values", stmt->lineNo );
	}

	free( tested );
}


void checkAssignmentStatement( Statement *stmt )
{
	checkVariableName( stmt->data.assignment_part.name );
	checkAssignmentExpression( stmt->data.assignment_part.kind, stmt->data.assignment_part.value );
}


void checkClassCallStatement( Statement *stmt )
{
	Member *member = stmt->data.call_part.definition.member;
	IdentList *arglist;
	ExpressionList *parmlist;

	/* Make sure the member has been linked and is either an output or a process */
	if( !member || ((member->kind != process_mbr) && (member->kind != output_mbr)) ) {
		semantic_error2( "%s.%s() not a process or output member.",
			stmt->data.call_part.name->lineNo,
			stmt->data.call_part.clazz->name,
			stmt->data.call_part.name->name );
		return;
	}

	for( arglist = member->data.normal_part.pArgs, parmlist = stmt->data.call_part.pArgs;
		arglist && parmlist; arglist = arglist->tail, parmlist = parmlist->tail )
	{
		switch( parmlist->head->kind )
		{
		case integer_expr:
		case string_expr:
			/* Nothing to check for integers */
			break;
		case constant_expr:
			checkConstantExpression( parmlist->head, FALSE, FALSE );
			break;
		default:
			/* Anything else is an error */
			semantic_error( "parameters must consist of only integers and constants.", parmlist->head->lineNo );
			break;
		}
	}

	/* Must be no arguments or parameters left over */
	if( arglist || parmlist ) {
		semantic_error2( "incorrect number of arguments to %s.%s().",
			stmt->data.call_part.name->lineNo,
			stmt->data.call_part.clazz->name,
			stmt->data.call_part.name->name );
		return;
	}
}


void checkStatement( Statement *stmt )
{
	switch( stmt->kind )
	{
	case if_stmt:
		checkConditionalExpression( stmt->data.if_part.condition, TRUE );
		checkStatementList( stmt->data.if_part.pThen );
		if( stmt->data.if_part.pElse )
			checkStatementList( stmt->data.if_part.pElse );
		break;
	case while_stmt:
	case do_stmt:
		checkConditionalExpression( stmt->data.while_and_do_part.condition, FALSE );
		checkStatementList( stmt->data.while_and_do_part.pBody );
		break;
	case for_stmt:
		checkAssignmentStatement( stmt->data.for_part.initial );
		checkConditionalExpression( stmt->data.for_part.condition, FALSE );
		checkAssignmentStatement( stmt->data.for_part.increment );
		checkStatementList( stmt->data.for_part.pBody );
		break;
	case switch_stmt:
		checkSwitchExpression( stmt->data.switch_part.expr );
		checkCaseBlockList( stmt->data.switch_part.pCaseBlocks );
		checkSwitchCoverage( stmt );
		break;
	case assignment_stmt:
		checkAssignmentStatement( stmt );
		break;
	case call_stmt:
		if( stmt->data.call_part.clazz )
			checkClassCallStatement( stmt );
		else
			semantic_error( "procedure calls not yet implemented.", stmt->data.call_part.name->lineNo );
		break;
	case return_stmt:
		semantic_error( "return not yet implemented.", stmt->data.call_part.name->lineNo );
		break;
	case break_stmt:
		semantic_error( "break (from loop) not yet implemented.", stmt->data.call_part.name->lineNo );
		break;
	default:
		semantic_error( "invalid statement.", stmt->lineNo );
		break;
	}
}


void checkStatementList( StatementList *stmtlist )
{
	for( ; stmtlist; stmtlist = stmtlist->tail )
	{
		if( stmtlist->head )
			checkStatement( stmtlist->head );
	}
}


void checkProcedure( Procedure *proc )
{
	checkStatementList( proc->pStmts );
}


void semanticCheck( Program *prog )
{
	PICList *piclist;
	ProcedureList *proclist;
	int mainfound = FALSE;

	for( piclist = prog->pPICs; piclist; piclist = piclist->tail )
	{
		if( piclist->head )
		{
			switch( piclist->head->kind )
			{
			case class_pic:
				if( piclist->head->data.clazz )
					checkClass( piclist->head->data.clazz );
				break;
			case instance_pic:
				if( piclist->head->data.instance )
					checkInstance( piclist->head->data.instance );
				break;
			case proc_pic:
				if( piclist->head->data.proc )
					checkProcedure( piclist->head->data.proc );
				break;
			}
		}
	}

	/* Check to see that there is a main proc */
	for( proclist = prog->pProcs; proclist; proclist = proclist->tail )
	{
		if( proclist->head && proclist->head->name )
		{
			if( strcmp( proclist->head->name->name, "main" ) == 0 )
			{
				mainfound = TRUE;
				break;
			}
		}
	}

	if( !mainfound )
		semantic_error( "Missing 'main' procedure.", NULL );
}
