/******************************************************************************
 *
 * Filename: transform.c
 *
 * Author:   Robert Thornburrow
 * Project:  Cycle - The CYbot Control LanguagE
 * Created:  7th June 2003
 *
 * Purpose:  Parse tree transformations - to make code easier to compile.
 *
 * (c) Copyright R.H.Thornburrow, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cycdefs.h"
#include "errors.h"


void transformAssignmentStatement( Statement* stmt )
{
	Expression* newvalue;

	/* Is it already an equals? */
	if( stmt->data.assignment_part.kind == equals_assign )
		return;

	switch( stmt->data.assignment_part.kind )
	{
	case plusequals_assign:
		newvalue = newBinaryOpExpression( addition_expr,
			newVariableExpr( copyIdent( stmt->data.assignment_part.name ) ),
			stmt->data.assignment_part.value );
		break;
	case minusequals_assign:
		newvalue = newBinaryOpExpression( subtraction_expr,
			newVariableExpr( copyIdent( stmt->data.assignment_part.name ) ),
			stmt->data.assignment_part.value );
		break;
	case timesequals_assign:
        newvalue = newBinaryOpExpression( multiplication_expr,
			newVariableExpr( copyIdent( stmt->data.assignment_part.name ) ),
			stmt->data.assignment_part.value );
		break;
	case divideequals_assign:
        newvalue = newBinaryOpExpression( division_expr,
			newVariableExpr( copyIdent( stmt->data.assignment_part.name ) ),
			stmt->data.assignment_part.value );
		break;
	case increment_assign:
        newvalue = newBinaryOpExpression( addition_expr,
			newVariableExpr( copyIdent( stmt->data.assignment_part.name ) ),
			stmt->data.assignment_part.value );
		break;
	case decrement_assign:
        newvalue = newBinaryOpExpression( subtraction_expr,
			newVariableExpr( copyIdent( stmt->data.assignment_part.name ) ),
			stmt->data.assignment_part.value );
		break;
	default:
		fprintf( stderr, "error: unkown assignment expression type.\n" );
		return;
	}

	/* Copy line number from existing expression */
	freeLineNumber( newvalue->lineNo );
	newvalue->lineNo = copyLineNumber( stmt->data.assignment_part.value->lineNo );

	/* Substitute the new expression for the old value */
	stmt->data.assignment_part.kind = equals_assign;
	stmt->data.assignment_part.value = newvalue;
}


void transformStatementList( StatementList** pStmtList )
{
	StatementList *stmtlist, *prev = NULL, *next, *temp;
	CaseBlockList *cblist;

	if( !*pStmtList )
		return;

	for( stmtlist = *pStmtList; stmtlist; prev = stmtlist, stmtlist = next )
	{
		next = stmtlist->tail;

		/* Remove NULL statements */
		if( !stmtlist->head )
		{
			semantic_warning( "warning: Removing null statement.", NULL );

			if( prev )
				prev->tail = stmtlist->tail;
			else
				*pStmtList = stmtlist->tail;

			free( stmtlist );
			stmtlist = prev;
			continue;
		}

		switch( stmtlist->head->kind )
		{
		case if_stmt:
			/* Transform the then and else bodies */
			transformStatementList( &stmtlist->head->data.if_part.pThen );
			transformStatementList( &stmtlist->head->data.if_part.pElse );
			break;
		case while_stmt:
		case do_stmt:
			/* Transform the loop body */
			transformStatementList( &stmtlist->head->data.while_and_do_part.pBody );
			break;
		case for_stmt:
			{
			Statement *forstmt = stmtlist->head;
			Statement *whilestmt;
			StatementList *incrpart;

			/* First expand any += and ++ etc. */
			transformAssignmentStatement( forstmt->data.for_part.initial );
			transformAssignmentStatement( forstmt->data.for_part.increment );

			/* Create a new while statement with the same condition and body as the for */
			whilestmt = newWhileStatement( forstmt->data.for_part.condition, forstmt->data.for_part.pBody );

			/* Move the line number of the for onto the while - free the one created with the while as it will be wrong */
			freeLineNumber( whilestmt->lineNo );
			whilestmt->lineNo = forstmt->lineNo;
			forstmt->lineNo = NULL;

			/* Null the condition and body of the for so that they are now owned by the while */
			forstmt->data.for_part.condition = NULL;
			forstmt->data.for_part.pBody = NULL;

			/* Substitute the while in place of the for in the statement list */
			stmtlist->head = whilestmt;

			/* Is there an initial assignment? */
			if( forstmt->data.for_part.initial )
			{
				/* Create a new StatementListEntry for the intial assigment and follow it with the while statement */
				temp = newStatementList( forstmt->data.for_part.initial, stmtlist ); 
				forstmt->data.for_part.initial = NULL;

				/* Insert the assignment in place of the for */
				if( prev )
					prev->tail = temp;
				else
					*pStmtList = temp;
			}

			/* Is there an increment? */
			if( forstmt->data.for_part.increment )
			{
				/* Create a new list element to contain the increment */
				incrpart = newStatementList( forstmt->data.for_part.increment, NULL );
				forstmt->data.for_part.increment = NULL;

				/* Tail insert the increment statement onto the body of the while statement */
				if( !whilestmt->data.while_and_do_part.pBody )
				{
					whilestmt->data.while_and_do_part.pBody = incrpart;
				}
				else
				{
					for( temp = whilestmt->data.while_and_do_part.pBody; temp->tail; temp = temp->tail );
					temp->tail = incrpart;
				}
			}

			/* Free the for statement */
			freeStatement( forstmt );
	
			/* Transform the loop body */
			transformStatementList( &whilestmt->data.while_and_do_part.pBody );
			}
			break;
		case switch_stmt:
			/* Transform each case block */
			for( cblist = stmtlist->head->data.switch_part.pCaseBlocks; cblist; cblist = cblist->tail )
				transformStatementList( &cblist->head->pStmts );
			break;
		case assignment_stmt:
			/* Expand any += and ++ etc. */
			transformAssignmentStatement( stmtlist->head );
			break;
		}
	}
}


void transformProcedure( Procedure* proc )
{
	if( !proc->pStmts )
		return;

	/* Insert a start statement at the beginning of main */
	if( strcmp( proc->name->name, "main" ) == 0 )
		proc->pStmts = newStatementList( newSimpleStatement( start_stmt ), proc->pStmts );

	transformStatementList( &proc->pStmts );
}


void transformCode( Program* prog )
{
	ProcedureList* proclist;

	for( proclist = prog->pProcs; proclist; proclist = proclist->tail )
	{
		if( proclist->head )
			transformProcedure( proclist->head );
	}
}
